#!/usr/bin/python

BASEDIR = "/home/nasroot/foscripts"
PROGRAM = "clean_settings"

import sys
import os

sys.path.append(os.path.join(BASEDIR, "bin/folib"))
import FailoverConfig
import HDIClient
import HDIOperation
from CommonLogger import log_debug, log_info, log_warning, log_error, log_exception, setup_rotation_logger
import Utility
from Utility import print_status, print_status_error

sys.path.append(os.path.join(BASEDIR, "bin"))
import Config
import Exceptions
import generate_sshkeys

def main():
    # Enable Ctrl-C interruption
    Utility.enable_interrupt()

    # Load configs
    config = Config.Config()

    # Setup logging
    log_file_path = os.path.join(config.log_dir, config.common_log_file)
    setup_rotation_logger(__name__, config.log_level, log_file_path, config.log_size, config.log_gen)

    log_info("----------------------- %s --------------------------" % PROGRAM)

    # Remove cron job
    # Need to execute as an external command because the script is signed.
    # No credential is required because the command is executed on localhost.
    print_status("Removing cron job ...")
    cred = HDIClient.Credential(None, None, None, None)
    conn = HDIOperation.Connection('localhost', cred, config.retry_count, config.retry_interval, config.timeout)
    prog = 'register_cron'
    cmd = "sudo maintexec %s --remove" % os.path.join(BASEDIR, prog)
    op = HDIOperation.CommandOperation(cmd)
    ret, out, err = op.execute(conn)
    if ret != 0:
        log_error("Failed to remove cron job. ret=%s" % ret)
        print_status_error("Failed to remove cron job. Stop execution.")
        raise Exceptions.OperationError("Failed to remove cron job.", ret)

    # Remove pair, pair group and storage policy configuration
    print_status("Removing pair, pair group and storage policy configuration ...")
    log_info("Getting list of pair configuration.")
    try:
        confgroup_list = FailoverConfig.list_configuration_group(config.fo_conf_dir)
    except FailoverConfig.ConfigurationError, e:
        log_error("Failed to count number of configurations due to an configuration error. error=%s" % str(e))
        raise Exceptions.ConfigurationError(str(e))

    for confgroup in confgroup_list:
        print_status("Removing configuration group %s ..." % confgroup)

        pairconf = FailoverConfig.PairConfig(config.fo_conf_dir, confgroup)
        pgconf = FailoverConfig.PairGroupConfig(config.fo_conf_dir, confgroup)
        stconf = FailoverConfig.StoragePolicyConfig(config.fo_conf_dir, confgroup)
        ssh_config = Config.SSHConfig(config.sys_conf_dir, pairconf, confgroup)

        Utility.remove_file(pairconf.config_file)
        Utility.remove_file(pgconf.config_file)
        Utility.remove_file(stconf.config_file)

        # Remove Unique SSH key for this configuration group
        print_status("Removing Unique SSH key for the configuration group ...")
        generate_sshkeys.remove_ssh_key_if_exists(ssh_config, is_default=False)

    # Register default SSH key
    ssh_config = Config.SSHConfig(config.sys_conf_dir)
    print_status("Registering default SSH key ...")
    generate_sshkeys.remove_ssh_key_if_exists(ssh_config, is_default=True)
    generate_sshkeys.register_ssh_key(ssh_config.get_public_key(is_default=True), is_default=True)

    print_status("Completed.")

if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        print "Interrupted."
        log_warning("Interrupted.")
        sys.exit(2)
    except Config.BaseError, e:
        print "Unable to load script configuration: %s" % str(e)
        sys.exit(10)
    except Exceptions.ConfigurationError, e:
        log_error("Unexpected configuration found: %s" % str(e))
        sys.exit(20)
    except Exceptions.OperationError, e:
        log_error("An operation error occurred: %s" % str(e))
        sys.exit(90)
    except EnvironmentError, e:
        log_exception("A system error occurred: %s", e)
        sys.exit(99)
    except Exception, e:
        log_exception("An internal error occurred: %s", e)
        sys.exit(99)

    sys.exit(0)
