#!/usr/bin/python
from __future__ import with_statement

BASEDIR = "/home/nasroot/foscripts"

import sys
import os
import errno
import re
import csv
import pickle

sys.path.append(os.path.join(BASEDIR, "bin/folib"))
from CommonLogger import log_debug, log_info, log_warning, log_error, log_exception
import Utility
from Utility import print_status
import HDIClient
import HDIOperation

sys.path.append(os.path.join(BASEDIR, "bin"))
import Exceptions

# Constants

job_dir = 'job'
result_dir = 'result'
result_file = 'self.csv'
storage_file = 'storage.pickle'

class ResultResource(object):
    PREP = 'PREP'
    FS = 'FS'
    DATA = 'DATA'
    CIFS = 'CIFS'
    NFS = 'NFS'
    TASK = 'TASK'
    OTHER = 'OTHER'

class RehearsalMode(object):
    DRYRUN = 1
    TRYOUT = 2
    def __init__(self, mode):
        if mode not in (self.DRYRUN, self.TRYOUT):
            raise Exceptions.ParameterError("Wrong rehearsal mode.")
        self.mode = mode

    def __str__(self):
        s = "(None)"
        if self.mode == self.DRYRUN:
            s = "Dry-run"
        elif self.mode == self.TRYOUT:
            s = "Try-out"
        return s

    def __eq__(self, other):
        return self.mode == other

    def __ne__(self, other):
        return self.mode != other

# Functions

def make_job_dir_path(config_dir):
    job_dir_path = os.path.join(config_dir, job_dir)
    try:
        os.mkdir(job_dir_path)
    except EnvironmentError, e:
        if e.errno != errno.EEXIST:
            log_error("Failed to make a system directory %s. error=%s" % (job_dir_path, str(e)))
            raise

    return job_dir_path

def make_result_dir_path(config_dir):
    result_dir_path = os.path.join(config_dir, result_dir)
    try:
        os.mkdir(result_dir_path)
    except EnvironmentError, e:
        if e.errno != errno.EEXIST:
            log_error("Failed to make a system directory %s. error=%s" % (result_dir_path, str(e)))
            raise

    return result_dir_path

def get_list_of_jobs(config_dir, is_error=False, is_success=False):
    """ Get a list of failover jobs that currently exist.

    This function lists up job files (, success job files or error job files
    which is specified) in config_dir and returns a list of the name of the jobs.

    Args:
        config_dir (str): Directory path of failover configuration being stored.
        is_error (bool): Specify True to get a list of 'error' jobs.
        is_success (bool): Specify True to get a list of 'success' jobs.

    Returns:
        list of str: List of job names.
    """

    job_dir_path = make_job_dir_path(config_dir)

    # Get a list of files under job directory
    try:
        file_list = os.listdir(job_dir_path)
    except EnvironmentError, e:
        log_error("Failed to list up files under %s. error=%s" % (job_dir_path, str(e)))
        raise

    # Get a list of job files
    if is_success:
        job_file_list = filter(is_success_job_file, file_list)
    elif is_error:
        job_file_list = filter(is_error_job_file, file_list)
    else:
        job_file_list = filter(is_job_file, file_list)
    log_debug("Job file list: %s" % job_file_list)

    if len(job_file_list) == 0:
        # No jobs exist
        job_name_list = []
    else:
        # Get a list of job names
        if is_success:
            job_name_list = map(get_name_of_success_job, job_file_list)
        elif is_error:
            job_name_list = map(get_name_of_error_job, job_file_list)
        else:
            job_name_list = map(get_name_of_job, job_file_list)

    return job_name_list

def get_next_job_name(config_dir, done_names=None, is_error=False, is_success=False):
    """ Return a name of a failover job that is supposed to be executed at the
    next.

    This function lists up job files (or error job files if is_error is True)
    in config_dir and returns the name of a job that is supposed to be executed
    at the next.

    Basically the order is dictionary order of name.

    Args:
        config_dir (str): Directory path of failover configuration being stored.
        done_names (list of str): List of failover job names that already processed.
            This function returns a name that is not contained in done_names if it's
            provided.
        is_error (bool): If this is true, this function finds and returns error job
            name.
        is_success (bool): If this is true, this function finds and returns success
            job name.

    Returns:
        str: Name of a failover job to be executed at the next.
            None is returned if no jobs are available.
    """

    if done_names is None:
        done_names = []

    job_name_list = get_list_of_jobs(config_dir, is_error=is_error, is_success=is_success)

    # Sort the list by name
    job_name_list.sort()
    log_debug("Sorted job name list: %s" % job_name_list)

    # Returns the first name which is not contained in done_names
    for job_name in job_name_list:
        if job_name not in done_names:
            name = job_name
            break
    else:
        # All jobs are done
        name = None

    return name

def create_job_file(name, source, fs, policy, config_dir):
    """ Create a new job file and record job information.

    This function creates a new job file with the specified name and records
    the provided job information to the file.

    This function expects that the directory to store a job file is already
    created.

    Args:
        name (str): A name for a failover job to create.
        source (str): IP address of the source node of this failover job.
        fs (str): File system name to be failed over by this job.
        policy (str): Storage policy for the file system. None can be
            specified.
        config_dir (str): Directory path to store failover configuration.
    Returns:
        None
    """
    # Make a job file name to create/write
    file_name = make_job_file_name(name)
    job_dir_path = make_job_dir_path(config_dir)
    job_file_path = os.path.join(job_dir_path, file_name)

    # Create and write the information with CSV format
    if os.path.exists(job_file_path):
        log_warning("A failover job for the same name already exists. Will be overwritten. name=%s, path=%s" % (name, job_file_path))

    try:
        with open(job_file_path, 'w') as f:
            writer = csv.writer(f, lineterminator='\n')
            if policy is None:
                writer.writerow([source, fs])
            else:
                writer.writerow([source, fs, policy])
    except EnvironmentError, e:
        log_error("Failed to create/write a failover job file. path=%s, error=%s" % (job_file_path, str(e)))
        raise

def read_job_file(name, config_dir, is_error=False, is_success=False):
    """ Read a job file and return the job information.

    Args:
        name (str): Name of a failover job to read.
        config_dir (str): Directory path to store failover configuration.
        is_error (bool): True when reading an 'error' job file.
        is_success (bool): True when reading an 'success' job file.
    Returns:
        Tuple (str, str, str):
            IP address of the source node,
            File system name,
            Storage policy string (can be None if not recorded)
    """
    # make a job file name to read
    if is_success:
        file_name = make_success_job_file_name(name)
    elif is_error:
        file_name = make_error_job_file_name(name)
    else:
        file_name = make_job_file_name(name)
    job_dir_path = make_job_dir_path(config_dir)
    job_file_path = os.path.join(job_dir_path, file_name)

    log_info("Reading job file %s ..." % file_name)

    # Read the job file
    try:
        with open(job_file_path, 'r') as f:
            reader = csv.reader(f)
            line = reader.next()
            if len(line) == 3:
                source = line[0]
                fs = line[1]
                policy = line[2]
            elif len(line) == 2:
                source = line[0]
                fs = line[1]
                policy = None
            else:
                log_error("Not enough / too many information recorded in a failover job file. path=%s, line=%s" % (job_dir_path, line))
                source = fs = policy = None
    except StopIteration:
        log_error("No lines are recorded in a faiover job file. path=%s" % job_dir_path)
        source = fs = policy = None
    except EnvironmentError, e:
        log_error("Failed to open/read a failover job file. path=%s, error=%s" % (job_file_path, str(e)))
        raise

    return (source, fs, policy)

def delete_job_file(name, config_dir, is_error=False, is_success=False):
    """ Delete a job file.

    This function deletes a job file which has the specified name of failover job.

    This function expects that a lock for failover job management is acquired.

    Args:
        name (str): Name of a failover job to be deleted.
        config_dir (str): Directory path to store failover configuration.
        is_error (bool): True when deleting an 'error' job file.
        is_success (bool): True when deleting an 'success' job file.
    Returns:
        None
    """
    # Make a job file name to delete
    if is_success:
        file_name = make_success_job_file_name(name)
    elif is_error:
        file_name = make_error_job_file_name(name)
    else:
        file_name = make_job_file_name(name)
    job_dir_path = make_job_dir_path(config_dir)
    job_file_path = os.path.join(job_dir_path, file_name)

    # Delete the file
    log_info("Deleting failover job file %s ..." % job_file_path)
    Utility.remove_file(job_file_path)

def mark_job_file_as_success(name, config_dir, is_error=False):
    """ Mark a failover job file as 'success' job file.

    This function marks a failover job as an 'success' job.
    'Successed' job file is kept for deciding which file system was successfully
    failed over by the latest failover activity.
    'Successed' job file is expected to be deleted by complete_failover command.

    Args:
        name (str): Name of a failover job to be marked as 'success'.
        config_dir (str): Directory path to store failover configuration.
        is_error (bool): True if marking an error job file as success.
    Returns:
        None
    """
    job_dir_path = make_job_dir_path(config_dir)

    # Job file path
    if is_error:
        file_name = make_error_job_file_name(name)
    else:
        file_name = make_job_file_name(name)
    job_file_path = os.path.join(job_dir_path, file_name)

    # Successed job file path
    success_file_name = make_success_job_file_name(name)
    success_job_file_path = os.path.join(job_dir_path, success_file_name)

    # Rename
    log_info("Marking a job as success. %s to %s." % (job_file_path, success_job_file_path))
    if os.path.exists(success_job_file_path):
        log_warning("An success failover job with the same name already exists. Will be overwritten. name=%s, path=%s" % (name, success_job_file_path))
    try:
        os.rename(job_file_path, success_job_file_path)
    except EnvironmentError, e:
        if e.errno == errno.ENOENT:
            # If the original job file is already gone.
            # There is nothing we can do. Just finish.
            log_info("Failover job file %s does not exist." % job_file_path)
        else:
            log_error("Rename a file failed. from=%s, to=%s" % (job_file_path, success_job_file_path))
            raise

def mark_job_file_as_error(name, config_dir):
    """ Mark a failover job file as 'error' job file.

    This function marks a failover job as an 'error' job.
    A job that is marked as 'error' is excluded from a list of jobs. A failover
    job where an error occurred needs to be excluded from a list of jobs,
    otherwise the job will be executed again immediately though it causes an error.
    We should not just delete an error job because we should have a chance to
    re-execute the job after the error is removed.

    Args:
        name (str): Name of a failover job to be marked as 'error'.
        config_dir (str): Directory path to store failover configuration.
    Returns:
        None
    """
    job_dir_path = make_job_dir_path(config_dir)

    # Job file path
    file_name = make_job_file_name(name)
    job_file_path = os.path.join(job_dir_path, file_name)

    # Error job file path
    error_file_name = make_error_job_file_name(name)
    error_job_file_path = os.path.join(job_dir_path, error_file_name)

    # Rename
    log_info("Marking a job as error. %s to %s." % (job_file_path, error_job_file_path))
    if os.path.exists(error_job_file_path):
        log_warning("An error failover job with the same name already exists. Will be overwritten. name=%s, path=%s" % (name, error_job_file_path))
    try:
        os.rename(job_file_path, error_job_file_path)
    except EnvironmentError, e:
        if e.errno == errno.ENOENT:
            # If the original job file is already gone.
            # There is nothing we can do. Just finish.
            log_info("Failover job file %s does not exist." % job_file_path)
        else:
            log_error("Rename a file failed. from=%s, to=%s" % (job_file_path, error_job_file_path))
            raise

def is_job_file(file_name, is_error=False, is_success=False):
    if is_success:
        m = re.match(r"failover_job_success\.[0-9a-zA-Z_]+", file_name)
    elif is_error:
        m = re.match(r"failover_job_error\.[0-9a-zA-Z_]+", file_name)
    else:
        m = re.match(r"failover_job\.[0-9a-zA-Z_]+", file_name)
    if m is None:
        return False
    else:
        return True

def is_error_job_file(file_name):
    return is_job_file(file_name, is_error=True)

def is_success_job_file(file_name):
    return is_job_file(file_name, is_success=True)

def get_name_of_job(file_name, is_error=False, is_success=False):
    if is_success:
        m = re.match(r"failover_job_success\.([0-9a-zA-Z_]+)", file_name)
    elif is_error:
        m = re.match(r"failover_job_error\.([0-9a-zA-Z_]+)", file_name)
    else:
        m = re.match(r"failover_job\.([0-9a-zA-Z_]+)", file_name)
    if m is None:
        return None
    else:
        return m.group(1)

def get_name_of_success_job(file_name):
    return get_name_of_job(file_name, is_success=True)

def get_name_of_error_job(file_name):
    return get_name_of_job(file_name, is_error=True)

def make_job_file_name(name):
    return "failover_job.%s" % name

def make_success_job_file_name(name):
    return "failover_job_success.%s" % name

def make_error_job_file_name(name):
    return "failover_job_error.%s" % name

def initialize_remote_result(target, config, ssh_config):
    log_info("Initializing failover results on %s ..." % target)

    # Setup connection to the target node
    cred = HDIClient.Credential(ssh_config.ssh_user, None, ssh_config.get_private_key(target), None)
    conn = HDIOperation.Connection(target, cred, config.retry_count, config.retry_interval, config.timeout)

    prog = os.path.join(BASEDIR, "int/initialize_result")
    cmd = "sudo maintexec %s" % prog
    op = HDIOperation.CommandOperation(cmd)
    try:
        ret, out, err = op.execute(conn)
    except HDIOperation.ConnectionError, e:
        log_error("Failed to connect %s over SSH. error=%s" % (target, str(e)))
        raise Exceptions.ConnectionError(e.reason, e.code)

    if ret != 0:
        log_error("Failed to initialize failover results on %s. error=%s" % (target, str(e)))
        raise Exceptions.OperationError(e.reason, e.code)

def initialize_result(config_dir):
    result_file_path = os.path.join(make_result_dir_path(config_dir), result_file)
    storage_file_path = os.path.join(make_result_dir_path(config_dir), storage_file)
    log_info("Removing failover results on this node (%s) ..." % result_file_path)
    Utility.remove_file(result_file_path)
    log_info("Removing temorary object on this node (%s) ..." % storage_file_path)
    Utility.remove_file(storage_file_path)

def save_storage(storage, config_dir):
    log_info("Saving storage information ...")
    storage_file_path = os.path.join(make_result_dir_path(config_dir), storage_file)
    try:
        with open(storage_file_path, 'wb') as f:
            pickle.dump(storage, f)
    except Exception, e:
        log_error("Failed to save storage information to a file. path=%s, error=%s" % (storage_file_path, str(e)))
        raise

def load_storage(config_dir):
    log_info("Loading storage information ...")
    storage = None
    storage_file_path = os.path.join(make_result_dir_path(config_dir), storage_file)
    try:
        with open(storage_file_path, 'rb') as f:
            storage = pickle.load(f)
    except EnvironmentError, e:
        if e.errno == errno.ENOENT:
            storage = None
        else:
            log_error("Failed to load storage information from a file. path=%s, error=%s" % (storage_file_path, str(e)))
            raise
    except Exception, e:
        log_error("Failed to load storage information from a file. path=%s, error=%s" % (storage_file_path, str(e)))
        raise

    return storage

def record_result_preparation(fsname, start, end, config_dir):
    log_info("Recording result of prepration ...")
    record_result(fsname, ResultResource.PREP, 1, start, end, config_dir)

def record_result_filesystem(fsname, start, end, config_dir):
    log_info("Recording result of file system restoration ...")
    record_result(fsname, ResultResource.FS, 1, start, end, config_dir)

def record_result_data(fsname, start, end, config_dir):
    log_info("Recording result of data restoration ...")
    record_result(fsname, ResultResource.DATA, 1, start, end, config_dir)

def record_result_cifs(fsname, num, start, end, config_dir):
    log_info("Recording result of CIFS shares restoration ...")
    record_result(fsname, ResultResource.CIFS, num, start, end, config_dir)

def record_result_nfs(fsname, num, start, end, config_dir):
    log_info("Recording result of NFS shares restoration ...")
    record_result(fsname, ResultResource.NFS, num, start, end, config_dir)

def record_result_other(fsname, start, end, config_dir):
    log_info("Recording result of other settings restoration ...")
    record_result(fsname, ResultResource.OTHER, 1, start, end, config_dir)

def record_result_task(fsname, num, start, end, config_dir):
    log_info("Recording result of migration tasks restoration ...")
    record_result(fsname, ResultResource.TASK, num, start, end, config_dir)

def record_result(fsname, resource, num, start, end, config_dir):
    result_file_path = os.path.join(make_result_dir_path(config_dir), result_file)

    record = [fsname, resource, num, start, end]
    try:
        with open(result_file_path, 'a') as f:
            writer = csv.writer(f, lineterminator='\n')
            writer.writerow(record)
    except EnvironmentError, e:
        log_error("Failed to create/write a failover result record file. path=%s, error=%s" % (result_file_path, str(e)))
        raise

def touch_result(config_dir):
    result_file_path = os.path.join(make_result_dir_path(config_dir), result_file)
    try:
        with open(result_file_path, 'a'):
            pass
    except EnvironmentError, e:
        log_error("Failed to create/open a failover result record file. path=%s, error=%s" % (result_file_path, str(e)))
        raise

def get_result(target, config_dir, config, ssh_config):
    """ Get failover result from target node.

    This function receives a result file recorded on the remote host specified
    by 'target', and saves it on a specific path on 'config_dir'. Then reads
    result records from the file and returns the result as a dictionary.

    This function raises an error if it fails to connect the target node.

    This function does not raise any errors even if there is an invalid record
    recorded in the received file. Instead returns result that it was able to
    read.

    Args:
        target (str): IP address of the node where failover result is received
            from.
        config_dir (str): Directory path to store failover result records.
        config (Config.Config): Configuration of the scripts.
        ssh_config (Config.SSHConfig): SSH configuration for the configuration
            group.

    Returns:
        dict: Failover result.
            {
                'fs01': {
                    'PREP': {
                        'num': Count of the resource
                        'start': Start time in UNIX time
                        'end': End time in UNIX time
                    },
                    'FS': {
                        ...
                    },
                    ...
                },
                ...
            }
    """
    # Setup connection to the target node
    cred = HDIClient.Credential(ssh_config.ssh_user, None, ssh_config.get_private_key(target), None)
    conn = HDIOperation.Connection(target, cred, config.retry_count, config.retry_interval, config.timeout)

    from_path = os.path.join(make_result_dir_path(config_dir), result_file)
    to_path = os.path.join(make_result_dir_path(config_dir), "%s.csv" % target)

    # Receive failover record file from the node
    transop = HDIOperation.TransferOperation(from_path, to_path)
    try:
        transop.receive(conn)
    except HDIOperation.ConnectionError, e:
        log_error("Failed to connect %s over SSH when receiving failover result data. error=%s" % (target, str(e)))
        # Return an empty result if we couldn't get result of the node, to display "No results" later.
        return {}

    # Read the file and construct dictionary
    result = {}
    try:
        with open(to_path, 'r') as f:
            reader = csv.reader(f)
            for items in reader:
                fsname = items[0]
                resource = items[1]
                num = int(items[2])
                start = int(items[3])
                end = int(items[4])
                if fsname not in result.keys():
                    result[fsname] = {}
                result[fsname][resource] = {'num': num, 'start': start, 'end': end}
    except (IndexError, ValueError), e:
        log_error("There are not enough values or invalid values in failover result record. error=%s" % str(e))
        # Pass through - return information that we could get anyway
    except EnvironmentError, e:
        log_error("Failed to open/read failover result file. path=%s, error=%s" % (to_path, str(e)))
        # Pass through - return information that we could get anyway

    return result

def display_result_line(node, fsname, resource, num, elapsed, elapsed_total, is_first_in_node=False, is_first_in_fs=False):
    if not is_first_in_node:
        node = ''
    if not is_first_in_fs:
        fsname = ''

    if resource is None:
        resource = '-'
    elif resource == ResultResource.PREP:
        resource = '(Preparation)'
    elif resource == ResultResource.FS:
        resource = 'File system'
    elif resource == ResultResource.DATA:
        resource = 'Data'
    elif resource == ResultResource.CIFS:
        resource = 'CIFS shares'
        resource += ' (x %s)' % num if num > 0 else ''
    elif resource == ResultResource.NFS:
        resource = 'NFS shares'
        resource += ' (x %s)' % num if num > 0 else ''
    elif resource == ResultResource.OTHER:
        resource = 'Other settings'
    elif resource == ResultResource.TASK:
        resource = 'Migration tasks'
        resource += ' (x %s)' % num if num > 0 else ''

    print "%-16s%-17s%-24s%-21s%s" % (node, fsname, resource, elapsed, elapsed_total)

def display_result(results, start, rehearsal=None):

    # Title
    if rehearsal is None:
        title = 'Failover results:'
        est = ''
    elif rehearsal == RehearsalMode.TRYOUT:
        title = 'Failover results (Try-out):'
        est = ''
    elif rehearsal == RehearsalMode.DRYRUN:
        title = 'Failover results (Dry-run):'
        est = ' (est.)'

    # Display title
    print
    print title
    print
    display_result_line('Node', 'File system', 'Restored resource', 0, 'Elapsed time' + est, 'Total elapsed time' + est, is_first_in_node=True, is_first_in_fs=True)
    print "--------------------------------------------------------------------------------------"
    for node in sorted(results.keys()):
        result_node = results[node]
        # First preparation time of a node is calculated as:
        #     start time of preparation of the first file system - start time of failover
        # Assumes file systems were restored in dictionary order of file system.
        fsnames = sorted(result_node.keys())
        if len(fsnames) == 0:
            # No file systems restored on the node? Just display "No results."
            display_result_line(node, '(No results)', None, 0, '-', '-', is_first_in_node=True, is_first_in_fs=True)
            print "--------------------------------------------------------------------------------------"
            continue
        fsname_first = fsnames[0]
        elapsed = result_node[fsname_first][ResultResource.PREP]['start'] - start
        elapsed_str = Utility.sec_to_hms(elapsed)
        display_result_line(node, '(Preparation)', None, 0, elapsed_str, elapsed_str, is_first_in_node=True, is_first_in_fs=True)
        last_elapsed_total = elapsed
        for fsname in sorted(result_node.keys()):
            is_first_in_fs = True
            result_fs = result_node[fsname]
            for resource in (ResultResource.PREP, ResultResource.FS, ResultResource.DATA, ResultResource.CIFS, ResultResource.NFS, ResultResource.OTHER, ResultResource.TASK):
                if resource in result_fs.keys():
                    result_resource = result_fs[resource]
                    num = result_resource['num']
                    elapsed = result_resource['end'] - result_resource['start']
                    elapsed_str = Utility.sec_to_hms(elapsed)
                    if rehearsal == RehearsalMode.DRYRUN:
                        elapsed_total = last_elapsed_total + elapsed
                    else:
                        elapsed_total = result_resource['end'] - start
                    elapsed_total_str = Utility.sec_to_hms(elapsed_total)
                    display_result_line(node, fsname, resource, num, elapsed_str, elapsed_total_str, is_first_in_fs=is_first_in_fs)
                else:
                    # No result information of the resource
                    display_result_line(node, fsname, resource, 0, '-', '-', is_first_in_fs=is_first_in_fs)
                last_elapsed_total = elapsed_total

                if is_first_in_fs:
                    is_first_in_fs = False
        print "--------------------------------------------------------------------------------------"
