#!/usr/bin/python
# All Rights Reserved. Copyright (c) 2018, Hitachi, Ltd.       #
# Licensed Material of Hitachi, Ltd.                           #
# Reproduction, use, modification or disclosure otherwise than #
# permitted in the License Agreement is strictly prohibited.   #
from __future__ import with_statement

import os
import sys
import fcntl
import logging
import logging.handlers
import inspect

# Logger instance
logger = None
lockfile = None

# Log level
class LogLevel(object):
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR

# Logging

def code_location(level=2):
    frame = inspect.currentframe(level)
    return "%s:%s" % (frame.f_code.co_name, frame.f_lineno)

def line_number(level=2):
    frame = inspect.currentframe(level)
    return frame.f_lineno

def filename(level=2):
    frame = inspect.currentframe(level)
    return os.path.basename(frame.f_code.co_filename)

def log_debug(msg, level=2):
    if logger is not None:
        with open(lockfile, 'w') as locker:
            fcntl.flock(locker, fcntl.LOCK_EX)
            logger.debug("%s: %s: %s" % (filename(level), code_location(level), msg))
            fcntl.flock(locker, fcntl.LOCK_UN)

def log_info(msg, level=2):
    if logger is not None:
        with open(lockfile, 'w') as locker:
            fcntl.flock(locker, fcntl.LOCK_EX)
            logger.info("%s:%s: %s" % (filename(level), line_number(level), msg))
            fcntl.flock(locker, fcntl.LOCK_UN)
    else:
        print >> sys.stderr, msg

def log_warning(msg, level=2):
    if logger is not None:
        with open(lockfile, 'w') as locker:
            fcntl.flock(locker, fcntl.LOCK_EX)
            logger.warning("%s:%s: %s" % (filename(level), line_number(level), msg))
            fcntl.flock(locker, fcntl.LOCK_UN)
    else:
        print >> sys.stderr, msg

def log_error(msg, level=2):
    if logger is not None:
        with open(lockfile, 'w') as locker:
            fcntl.flock(locker, fcntl.LOCK_EX)
            logger.error("%s:%s: %s" % (filename(level), line_number(level), msg))
            fcntl.flock(locker, fcntl.LOCK_UN)
    else:
        print >> sys.stderr, msg

def log_exception(msg, e, level=2):
    if logger is not None:
        with open(lockfile, 'w') as locker:
            fcntl.flock(locker, fcntl.LOCK_EX)
            logger.exception("%s:%s: %s" % (filename(level), line_number(level), msg), e)
            fcntl.flock(locker, fcntl.LOCK_UN)
    else:
        print >> sys.stderr, "%s: %s" % (msg, str(e))

def setup_rotation_logger(name, log_level, log_file_path, log_size, log_gen, print_msg=False):
    # Configure root logger so that other modules can use the same logging configuration
    global logger
    logger = logging.getLogger(name)
    filehandler = logging.handlers.RotatingFileHandler(log_file_path, 'a', log_size * 1024 * 1024, log_gen)
    consolehandler = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s: %(process)d: %(levelname)8s: %(message)s')
    filehandler.setFormatter(formatter)
    consolehandler.setFormatter(formatter)

    # Log level needs to be set to both logger and handler
    # Log level of logger is checked at first. Then log level of handler is checked.

    # Default log level
    logger.setLevel(logging.INFO)
    filehandler.setLevel(logging.DEBUG) # Output all log levels that logger allows
    consolehandler.setLevel(logging.WARNING)

    if not log_level in (logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR):
        # A wrong log level falls into INFO
        log_level = logging.INFO

    if log_level in (logging.DEBUG, logging.INFO):
        logger.setLevel(log_level)
        if print_msg:
            consolehandler.setLevel(log_level)
    else:
        logger.setLevel(log_level)
        consolehandler.setLevel(log_level)
    logger.addHandler(filehandler)
    logger.addHandler(consolehandler)

    # Setup lock file for exclusive logging to the same log file
    log_dir = os.path.dirname(log_file_path)
    lock_file = os.path.basename(log_file_path) + ".lck"
    lock_dir = os.path.join(log_dir, "lock")
    if not os.path.exists(lock_dir):
        os.mkdir(lock_dir)
    global lockfile
    lockfile = os.path.join(lock_dir, lock_file)

def setup_stdout_logger(name, log_level):
    global logger
    logger = logging.getLogger(name)
    consolehandler = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s: %(process)d: %(levelname)8s: %(message)s')
    consolehandler.setFormatter(formatter)

    if not log_level in (logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR):
        # A wrong log level falls into INFO
        log_level = logging.INFO
    consolehandler.setLevel(log_level)
    logger.setLevel(log_level)
    logger.addHandler(consolehandler)
