#!/usr/bin/python
# All Rights Reserved. Copyright (c) 2018, Hitachi, Ltd.       #
# Licensed Material of Hitachi, Ltd.                           #
# Reproduction, use, modification or disclosure otherwise than #
# permitted in the License Agreement is strictly prohibited.   #

import sys
import os
import subprocess
import time
import datetime
import string
import random
import distutils.util
import signal
import socket
import getpass
import errno
import re

try:
    from CommonLogger import log_debug, log_info, log_warning, log_error
except ImportError:
    # To make this module handy and portable
    def log_debug(msg): pass
    def log_info(msg): pass
    def log_warning(msg): pass
    def log_error(msg): pass

# Global variables
ID = None       # ID is added to messages displayed by print_status() if it's defined

# Class

class Version(object):
    """ Class to represent a version to enable version comparison.
    """
    def __init__(self, version):
        # version must be a string with numbers split by '.', like '6.4.2.13'
        nums = version.split('.')
        self.major = int(nums[0])
        self.minor = int(nums[1]) if len(nums) > 1 else 0
        self.revision = int(nums[2]) if len(nums) > 2 else 0
        self.build = int(nums[3]) if len(nums) > 3 else 0

    def __lt__(self, other):
        return self._cmp(other) == -1

    def __le__(self, other):
        return self._cmp(other) == -1 or self._cmp(other) == 0

    def __eq__(self, other):
        return self._cmp(other) == 0

    def __ne__(self, other):
        return self._cmp(other) != 0

    def __gt__(self, other):
        return self._cmp(other) == 1

    def __ge__(self, other):
        return self._cmp(other) == 1 or self._cmp(other) == 0

    def __str__(self):
        return "%d.%d.%d.%d" % (self.major, self.minor, self.revision, self.build)

    def _cmp(self, other):
        """ Compare two versions.

        Return -1 if self < other.
        Return 0 if self == other.
        Return 1 if self > other.
        """
        major_diff = self.major - other.major
        minor_diff = self.minor - other.minor
        revision_diff = self.revision - other.revision
        build_diff = self.build - other.build

        for diff in (major_diff, minor_diff, revision_diff, build_diff):
            if diff > 0:
                result = 1
                break
            elif diff < 0:
                result = -1
                break
            # Go to the next diff if diff == 0
        else:
            # All versions are same
            result = 0

        return result

class Color:
    BLACK = '\033[30m'
    RED = '\033[31m'
    GREEN = '\033[32m'
    YELLOW = '\033[33m'
    BLUE = '\033[34m'
    PURPLE = '\033[35m'
    CYAN = '\033[36m'
    WHITE = '\033[37m'
    END = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    INVISIBLE = '\033[08m'
    REVERSE = '\033[07m'

# Command

def run(command, display_stdout=False):
    p = subprocess.Popen(['sh', '-c', command], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=False)

    lines = ""
    while True:
        line = p.stdout.readline()
        if line:
            lines += line
            if display_stdout:
                sys.stdout.write(line)
        else:
            break

    # Not expecting stderr content is large
    err = ''.join(p.stderr.readlines())
    ret = p.wait()

    return ret, lines, err

# Question

def yes_or_no(message):
    while True:
        answer = my_input(message + " [y/n]: ").lower()
        log_info("Question: %s, answer: %s" % (message, answer))
        if answer in ["y", "yes"]:
            return True
        elif answer in ["n", "no"]:
            return False

def yes_or_no_full(message):
    while True:
        answer = my_input(message + " [yes/no]: ").lower()
        log_info("Question: %s, answer: %s" % (message, answer))
        if answer == "yes":
            return True
        elif answer == "no":
            return False

def yes_or_no_full_confirm(message):
    while True:
        answer = my_input(message + " [yes/no]: ").lower()
        log_info("Question: %s, answer: %s" % (message, answer))
        if answer == "yes":
            message2 = "Is your answer \"yes\"?"
            answer2 = my_input(message2 + " [yes/no]: ").lower()
            log_info("Question: %s, answer: %s" % (message2, answer2))
            if answer2 == "yes":
                return True
            elif answer2 == "no":
                return False
        elif answer == "no":
            return False

# Utilities

def my_input(msg):
    try:
        answer = raw_input(msg)
    except EOFError:
        log_info("Nothing is input.")
        print "Nothing is input."
        answer = None

    return answer

def my_input_secret(msg):
    try:
        answer = getpass.getpass(msg)
    except EOFError:
        log_info("Nothing is input.")
        print "Nothing is input."
        answer = None

    return answer

def is_word(word):
    # Checks if the provided 'word' consists of alphanumeric characters and underscore.
    m = re.match(r"^[a-zA-Z0-9_]+$", word)
    if m is None:
        return False
    else:
        return True

def _print_status(msg, yellow=False, red=False):
    # Display the message with date, time and hostname
    ret, out, err = run('date "+%Y-%m-%d %H:%M:%S %Z"')
    now = out.strip()
    hostname = os.uname()[1]
    id_msg = "[ID: %s] " % ID if ID is not None else ""

    # Highlight with color if specified
    if yellow:
        msg = Color.REVERSE + Color.YELLOW + msg + Color.END
    elif red:
        msg = Color.REVERSE + Color.RED + msg + Color.END

    # Print
    msg = id_msg + msg
    print "%s: %s: %s" % (now, hostname, msg)

def print_status(msg):
    # Log the same message
    log_info(msg, level=3)
    _print_status(msg)

def print_status_warning(msg):
    # Log the same message
    log_warning(msg, level=3)
    _print_status(msg, yellow=True)

def print_status_error(msg):
    # Log the same message
    log_error(msg, level=3)
    _print_status(msg, red=True)

def str2bool(s):
    return bool(distutils.util.strtobool(s))

def bool2str(boolean):
    return str(boolean).lower()

def sec_to_hms(sec):
    if sec < 0:
        negative = True
        sec = -sec
    else:
        negative = False
    m, s = divmod(sec, 60)
    h, m = divmod(m, 60)

    if negative:
        hms = "-%02d:%02d:%02d" % (h, m, s)
    else:
        hms = "%02d:%02d:%02d" % (h, m, s)
    return hms

def epoch_to_iso8601(epoch):
    # Convert epoch time to ISO 8601 format.
    # Timezone on the node is used. (If this is executed on a node on EDT, timezone string will be "-04:00".

    # --rfc-3339 option can't be used here because it doesn't have 'millisecond' unit...
    cmd = "date +%%Y-%%m-%%dT%%T.000%%:z -d @%s" % epoch
    log_debug("Command-Exec-Start: %s" % cmd)
    ret, out, err = run(cmd)
    log_debug("Command-Exec-End: %s, ret=%s" % (cmd, ret))
    if ret != 0:
        log_error("date command failed. ret=%s, out=%s, err=%s" % (ret, out, err))
        return None

    return out.strip()

def iso8601_to_epoch(isotime):
    # Convert ISO 8601 format time string to epoch (integer).

    isotime = isotime.replace('T', ' ')
    cmd = 'date -d "%s" +%%s' % isotime
    log_debug("Command-Exec-Start: %s" % cmd)
    ret, out, err = run(cmd)
    log_debug("Command-Exec-End: %s, ret=%s" % (cmd, ret))
    if ret != 0:
        log_error("date command failed. ret=%s, out=%s, err=%s" % (ret, out, err))
        return None

    return int(out)

def interrupt_handler(sig, frame):
    raise KeyboardInterrupt

def enable_interrupt():
    # Accept SIGINT.
    # This function has to be called in a signed script because maintexec blocks SIGINT.
    signal.signal(signal.SIGINT, interrupt_handler)

def deep_str(obj):
    # Convert obj to string, with converting each child of obj
    if isinstance(obj, dict):
        result = {}
        for k, v in obj.items():
            result[k] = str(v)
    elif hasattr(obj, '__iter__'):
        result = []
        for e in obj:
            if hasattr(e, '__iter__'):
                e_str = map(str, e)
                result.append(str(e_str))
            else:
                result.append(str(e))
    else:
        result = obj

    return str(result)

def convert_to_time_on_HDI(epoch_msec, time_diff):
    epoch = epoch_msec / 1000
    # Adjust HCP time to HDI time
    epoch = epoch - time_diff
    date = datetime.datetime.fromtimestamp(epoch)

    return date

def check_selected_num(nums, min_count, max_count, onlyone=False):
    if len(nums) == 0:
        print "Nothing was selected"
        log_info("Nothing was selected")
        return False

    if nums[0] == "":
        print "Nothing was selected"
        log_info("Nothing was selected")
        return False

    if onlyone and len(nums) > 1:
        print "Only one item can be specified."
        log_info("Only one item can be specified.")
        return False

    nums_uniq = list(set(nums))
    if len(nums_uniq) != len(nums):
        print "Same value is specified multiple times."
        log_warning("Same value is specified multiple times.")
        return False

    for num in nums:
        if not num.isdigit() or (int(num) < min_count or int(num) > max_count):
            print "An invalid value is specified. Select between %s to %s." % (min_count, max_count)
            log_warning("An invalid value is specified. Select between %s to %s." % (min_count, max_count))
            return False

    return True

def select_items(msg, menu, title=None, onlyone=False, allow_empty=False):
    retry = 2
    tried = 0

    log_info("Menu: %s" % menu)

    # Display menu
    print
    display_menu(menu, title)
    print

    # Include selectable number range in the message
    end = -1
    start = sys.maxint
    for idx, value in menu:
        # idx can be a string
        if not str(idx).isdigit():
            continue
        idxnum = int(idx)
        if idxnum > end:
            end = idxnum
        if idxnum < start:
            start = idxnum
    if start == end:
        msg = msg + " [%s]: " % (start)
    else:
        msg = msg + " [%s-%s]: " % (start, end)

    # Select an item(s)
    while tried <= retry:
        tried += 1
        # Get user's input
        answer = my_input(msg)
        if answer is None:
            # EOF is input
            if allow_empty:
                result = []
            else:
                result = None
            break

        log_info("Input value: %s" % answer)

        if len(answer) == 0 and allow_empty:
            result = []
            break

        nums = answer.split(',')
        # Check input value
        if not check_selected_num(nums, start, end, onlyone):
            continue
        result = map(int, nums)
        break
    else:
        result = None

    return result

def display_menu(menu, title=None):
    # menu is a list of tuple (index, value)

    if not isinstance(menu, list):
        raise TypeError("Non dictionary data is passed.")

    if len(menu) == 0:
        raise ValueError("No items for the menu.")

    if title is not None:
        print title + ":"
    for item in menu:
        if not isinstance(item, tuple):
            raise ValueError("Item is not a tuple.")
        idx, value = item
        print "\t%s. %s" % (idx, value)

def is_IP_address(ipaddr):
    # Check the provided string satisfies IP address format

    if not isinstance(ipaddr, str):
        return False

    try:
        socket.inet_pton(socket.AF_INET, ipaddr)
    except socket.error, e:
        log_debug("IP address format check: %s" % str(e))
        return False
    return True

def safe_remove_from_list(l, e):
    """ Remove an element in a list if it exists.
    l: List
    e: Element to be removed
    """
    if e in l:
        l.remove(e)

def remove_file(file_path):
    try:
        os.remove(file_path)
    except EnvironmentError, e:
        if e.errno == errno.ENOENT:
            log_info("A file %s is already deleted." % file_path)
        else:
            log_error("Failed to delete a file %s. error=%s" % (file_path, str(e)))
            raise

def get_list_of_files(dir_path):
    try:
        file_list = os.listdir(dir_path)
    except EnvironmentError, e:
        log_error("Failed to get a list of files on %s. error=%s" % (dir_path, str(e)))
        raise

    return file_list

def history_to_unixtime(history_name, time_diff):
    dt = datetime.datetime.strptime(history_name, "%Y_%m_%d_%H%M")
    utime = int(time.mktime(dt.timetuple())) + time_diff
    return utime

def history_to_timestring(history_name):
    try:
        dt = datetime.datetime.strptime(history_name, "%Y_%m_%d_%H%M")
    except ValueError:
        return "(null)"
    return "%4d-%02d-%02d %02d:%02d" % (dt.year, dt.month, dt.day, dt.hour, dt.minute)

def human_readable_size(size):
    units = ["KB", "MB", "GB", "TB"]
    multiple = 1024

    if size is None:
        return "None"
    for unit in units:
        size = float(size) / multiple
        if size < multiple:
            return "%.1f %s" % (size, unit)

def random_string(num):
    return ''.join([random.choice(string.digits + string.letters) for i in range(num)])

def is_expected_HTTP_status(ret, expect_code):
    """ Check errors of HTTP return code.

    Returns False if the returned HTTP code does not match any expected codes.
    Otherwise returns True.

    TypeError or KeyError is raised if the values of arguments are wrong.

    Args:
        ret (dict): A dictionary returned by any REST method of HCPClient.
        expect_code (tuple of int): A list of HTTP status codes to be expected.
    Returns:
        bool: Returns True if the return code is expected. Otherwise False.
    """

    if ret['code'] in expect_code:
        return True
    else:
        if 'X-HCP-ErrorMessage' in ret['headers']:
            errmsg = ret['headers']['X-HCP-ErrorMessage']
        else:
            errmsg = "None"
        log_error("Unexpected HTTP status code. ret=%s, msg=%s, error=%s" % (ret['code'], ret['message'], errmsg))
        return False
