#!/usr/bin/python

BASEDIR = "/home/nasroot/foscripts"
PROGRAM = "setup_storage_policy"

import sys
import os

sys.path.append(os.path.join(BASEDIR, "bin/folib"))
import FailoverConfig
import HDIConfig
from CommonLogger import log_debug, log_info, log_warning, log_error, log_exception, setup_rotation_logger
import Utility

sys.path.append(os.path.join(BASEDIR, "bin"))
import Config
import Exceptions
import setup_common
import getinfo_common

# Global variables which save information of a remote node,
# to reduce count of external access.
# These are dictionary to store information for each node. Key is IP address of the node.
_model = {}
_capacity = {}
_lulist = {}

# Accessor of the global variables
def get_remote_model(ipaddr, config, ssh_config):
    global _model
    if ipaddr not in _model:
        _model[ipaddr] = getinfo_common.get_remote_model(ipaddr, config, ssh_config)

    return _model[ipaddr]

def get_remote_available_capacity(ipaddr, config, ssh_config):
    global _capacity
    if ipaddr not in _capacity:
        _capacity[ipaddr] = getinfo_common.get_remote_available_capacity(ipaddr, config, ssh_config)

    return _capacity[ipaddr]

def get_remote_available_lulist(ipaddr, config, ssh_config):
    # TODO: Need to consider external storage case? LU name can conflict?
    global _lulist
    if ipaddr not in _lulist:
        _lulist[ipaddr] = getinfo_common.get_remote_available_lulist(ipaddr, config, ssh_config)

    return _lulist[ipaddr]

def show_policy_help(model, available_policies):
    print
    print "--- Help ---"
    print
    for i, policy in enumerate(available_policies):
        idx = i + 1
        print "%s. %s" % (idx, FailoverConfig.StoragePolicy.policy_to_display_string(policy))
        if model == HDIConfig.Model.CLUSTER:
            print FailoverConfig.StoragePolicy.policy_to_help_string_cluster(policy)
        elif model == HDIConfig.Model.SINGLE:
            print FailoverConfig.StoragePolicy.policy_to_help_string_single(policy)

    print "------------"

def select_policy(model, is_default=False, is_all=False):
    """ Select a policy from a list of selectable policies with the specified condition.

        This function displays a list of selectable storage policies and let a user
        select one of them from standard input. Then returns the selected policy as
        StoragePolicy object.
        Selectable policies vary depending on the condition that is determined by
        the arguments.
        This function just selects a policy constant. Setting additional information
        for some type of policy is caller's responsibility.

        Example:
            'Like for Like' is selectable only as a default policy (is_default is
            True) AND all file systems are to be failed over (is_all is True).
            'Manual' is not selectable as default policy (is_default is False).

        Args:
            model (int): A constant that represents the model of the secondary node.
            is_default (bool): True when setting up a default policy.
            is_all (bool): True if all file systems are to be failed over.

        Returns:
            int: A storage policy constant selected by a user.
    """

    # Get all defined policies
    available_policies = FailoverConfig.StoragePolicy.get_available_policies()

    if is_default:
        # Manual is not available as default policy
        Utility.safe_remove_from_list(available_policies, FailoverConfig.StoragePolicy.MANUAL)
    if not is_default or not is_all:
        # Like for Like is available as default policy AND all file systems are to be failed over
        Utility.safe_remove_from_list(available_policies, FailoverConfig.StoragePolicy.LIKEFORLIKE)

    if model == HDIConfig.Model.SINGLE:
        # Auto - pickup is selectable only for cluster
        Utility.safe_remove_from_list(available_policies, FailoverConfig.StoragePolicy.AUTOPICKUP)

    # Build a menu
    menu = []
    menu_idx_map = {}
    for i, value in enumerate(available_policies):
        idx = i + 1
        menu.append((idx, FailoverConfig.StoragePolicy.policy_to_display_string(value)))
        menu_idx_map[idx] = value
    # Add help menu
    menu.append((0, 'Show help'))

    # Select a policy from the menu.
    # Keep asking until other than 'help' is selected.
    while True:
        selected_list = Utility.select_items("Select a policy", menu, title="Selectable storage policies", onlyone=True)
        if selected_list is None:
            log_warning("Selecting a storage policy did not succeed.")
            print "Process canceled."
            raise Exceptions.NoExecution()
        selected_idx = selected_list[0]
        if selected_idx == 0:
            show_policy_help(model, available_policies)
            print
            print "Going back to policy selection menu ..."
        else:
            break

    selected_policy = menu_idx_map[selected_idx]
    print
    print "'%s' is selected." % FailoverConfig.StoragePolicy.policy_to_display_string(selected_policy)

    return selected_policy

def input_number(question, allow_empty=False):
    """ Get a number input from stdin.

    Ask a user to input a number with a question message specified with the
    first argument.
    Raise an exception when nothing is input, just enter is input or non-number
    is input.
    Returns the number as an integer value.

    Args:
        question (str): A question message (prompt) to be displayed when asking
            to input a number.
        allow_empty (bool): If this is True, this function returns None if
            just enter is input. Otherwise this function raises an exception
            in the same situation.
    Returns:
        int: An integer value that was input from stdin.
    """
    number = Utility.my_input(question)
    if number is None:
        # EOF
        raise Exceptions.NoExecution()

    if len(number) == 0:
        # Just enter
        if allow_empty:
            return None
        else:
            raise Exceptions.NoExecution()

    if not number.isdigit():
        print "Input data is not a number."
        raise Exceptions.NoExecution()

    return int(number)

def input_margin_capacity():
    print
    print ">>> Specify a 'margin' capacity to be added to the usage of the file system."
    print

    capacity = input_number("Capacity (GB) (Just press enter to specify 'no margin'): ", allow_empty=True)
    log_info("Margin: %s" % capacity)
    if capacity is None:
        capacity = 0

    return capacity

def input_specified_capacity():
    print
    print ">>> Specify a (minimum) capacity to be assigned for each file system."
    print

    capacity = input_number("Capacity (GB): ")
    log_info("Specified capacity: %s" % capacity)

    if capacity < 1:
        print "The specified capacity is too small."
        raise Exceptions.NoExecution()

    return capacity

def input_manual_storage_configuration(fs, model, is_actmig, pair, config, ssh_config):
    """ Setup a manual policy.

    For cluster:
        Get list of available LUs on the secondary node, and let user select
        LU(s) for the file system.
    For single/VM:
        Get available VG capacity on the secondary node, and let user input
        capacity to assign to the file system.

    If Active Migration is enabled on the file system, ask to input LUs/capacity
    for work space.
    Build StoragePolicy object according to the entered information and return it.

    Args:
        fs (str): File system name for this manual policy.
        model (int): Model of the secondary node.
        pair (FailoverConfig.Pair): Pair object of this pair.
        config (Config.Config): Configuration object of the scripts.
        ssh_config (Config.SSHConfig): SSH configuration.

    Returns:
        FailoverConfig.StoragePolicy: Storage policy object.
    """

    if model == HDIConfig.Model.CLUSTER:
        # Get available LUs on the secondary node
        lu_list = get_remote_available_lulist(pair.secondary, config, ssh_config)

        # Select an LU(s) for the file system
        print
        print ">>> Select an LU(s) to assign to %s." % fs
        selected_list = select_lu(lu_list)
        fslu_list = [lu_list[idx].deviceFileName for idx in selected_list]

        # Select an LU(s) for the work space if Active Migration is enabled
        if is_actmig:
            print
            print ">>> Select an LU(s) to assign to %s's work space." % fs
            selected_list = select_lu(lu_list)
            wslu_list = [lu_list[idx].deviceFileName for idx in selected_list]
        else:
            wslu_list = []

        # Check LU conflicts
        fsluset = set(fslu_list)
        wsluset = set(wslu_list)
        if len(fsluset & wsluset) != 0:
            print "The same LU is specified for both file system and work space."
            log_warning("The same LU is specified for both file system and work space. file system: %s, work space: %s" % (fslu_list, wslu_list))
            raise Exceptions.NoExecution()

        # Build a StoragePolicy instance
        policy = FailoverConfig.StoragePolicy(fs, policy=FailoverConfig.StoragePolicy.MANUAL, fslulist=fslu_list, wslulist=wslu_list)

    elif model == HDIConfig.Model.SINGLE:
        # Get available VG capacity on the secondary node
        free = get_remote_available_capacity(pair.secondary, config, ssh_config)
        print
        print "Available capacity on %s: %s (GB)" % (pair.secondary, free)

        # Specify a capacity for the file system
        print
        print ">>> Specify a capacity to assign to %s." % fs
        fs_capacity = input_number("Capacity (GB): ")
        log_info("Capacity for file system %s: %s" % (fs, fs_capacity))

        # Specify a capacity for the work space if Active Migration is enabled
        if is_actmig:
            print
            print ">>> Specify a capacity to assign to %s's work space." % fs
            ws_capacity = input_number("Capacity (GB) (Just press enter to calculate automatically): ", allow_empty=True)
            log_info("Capacity for file system %s: %s" % (fs, ws_capacity))
        else:
            ws_capacity = None

        # Build a StoragePolicy instance
        policy = FailoverConfig.StoragePolicy(fs, policy=FailoverConfig.StoragePolicy.MANUAL, fs_capacity=fs_capacity, ws_capacity=ws_capacity)

    return policy

def select_lu(lu_list):
    """ Select an LU(s) with menu style.

    Display the list of LUs specified by lu_list and let a user select one or
    more of them.
    Returns index(es) of lu_list as a list.

    Args:
        lu_list (list of HDIConfig.LUInfo): A list of LUs that a user selects from.

    Returns:
        List of int: A list of index numbers of lu_list that were selected by
        a user.
    """
    lu_menu = []
    for i, lu in enumerate(lu_list):
        idx = i + 1
        lu_menu.append((idx, "%s - %s GB" % (lu.deviceFileName, lu.size)))
    selected_list = Utility.select_items("Select an LU(s)", lu_menu, title="Available LUs on the secondary node")
    if selected_list is None:
        log_warning("Selecting an LU did not succeed.")
        print "Process canceled."
        raise Exceptions.NoExecution()

    return [x - 1 for x in selected_list]

def change_default_policy(pid, pair, stconf, config, ssh_config):
    """ Change default policy for a pair.

    Basically same as define_policy() except for the followings:
    - Some policies that cannot be configured as default are not selectable.
    - User does not select any file systems.

    Args:
        pid (int): Pair ID of the pair.
        pair (FailoverConfig.Pair): Pair object of the pair.
        stconf (FailoverConfig.StoragePolicyConfig): Storage policy configuration
            object to be updated.
        config (Config.Config): Configuration object of the script.
        ssh_config (Config.SSHConfig): SSH configuration.

    Returns:
        None: Returns nothing. All changes are updated in stconf.
     """

    print
    print "----- Change the default policy -----"

    # Get model information of the secondary node
    model = get_remote_model(pair.secondary, config, ssh_config)

    # Select a policy
    policy_id = select_policy(model, is_default=True, is_all=pair.is_all)

    # Build StoragePolicy.
    # Some additional information is required for some policy types. Do not need to
    # care about 'Manual' policy because it is never selected for default policy.
    if policy_id == FailoverConfig.StoragePolicy.AUTOUSAGE:
        margin = input_margin_capacity()
        policy = FailoverConfig.StoragePolicy(FailoverConfig.StoragePolicy.FSDEF, policy=policy_id, margin=margin)
    elif policy_id == FailoverConfig.StoragePolicy.AUTOSPEC:
        capacity = input_specified_capacity()
        policy = FailoverConfig.StoragePolicy(FailoverConfig.StoragePolicy.FSDEF, policy=policy_id, spec=capacity)
    else:
        policy = FailoverConfig.StoragePolicy(FailoverConfig.StoragePolicy.FSDEF, policy=policy_id)

    # Delete and add default policy from/to storage configuration
    stconf.delete(pid, target_fs=FailoverConfig.StoragePolicy.FSDEF)
    stconf.add(pid, policy=policy)

def define_policy(pid, pair, stconf, config, ssh_config):
    """ Define a storage policy for a file system(s).

    Define a storage policy for selected file system(s).
    User selects file systems to configure storage policies at first.
    Then select a policy for the selected file systems, with specifying additional
    information if needed.

    Args:
        pid (int): Pair ID of the pair.
        pair (FailoverConfig.Pair): Pair object of the pair.
        stconf (FailoverConfig.StoragePolicyConfig): Storage policy configuration
            object to be updated.
        config (Config.Config): Configuration object of the script.
        ssh_config (Config.SSHConfig): SSH configuration.

    Returns:
        None: Returns nothing. All changes are updated in stconf.
    """

    print
    print "----- Define a policy -----"
    # Get model information of the secondary node
    model = get_remote_model(pair.secondary, config, ssh_config)

    # Get a list of file systems on the primary node
    fs_list = getinfo_common.get_remote_file_systems(pair.primary, config, ssh_config)
    if len(fs_list) == 0:
        log_warning("No file systems found on the primary node (%s)." % pair.primary)
        print "No file systems found on the primary node (%s). Please configure a default policy at the moment."
        raise Exceptions.NoExecution()

    # Remove file systems that will not be failed over from fs_list
    if not pair.is_all:
        for fs in fs_list:
            if fs.fileSystemName not in pair.target_fslist:
                print "%s is not selected as failover target." % fs.fileSystemName
            elif not fs.is_supported():
                print "%s is not supported for failover." % fs.fileSystemName
        fs_list = filter(lambda fs: fs.fileSystemName in pair.target_fslist, fs_list)
        fs_list = [fs for fs in fs_list if fs.is_supported()]

    # Select a file system(s) to setup a policy
    print
    print ">>> Specify a file system(s) to setup a storage policy."
    fs_menu = []
    for i, fs in enumerate(fs_list):
        idx = i + 1
        fs_menu.append((idx, "%s - %s GB" % (fs.fileSystemName, fs.fileSystemSizeInGigaByte)))
    selected_list = Utility.select_items("Select a file system(s) (separate with comma for multiple selection)", fs_menu, title="File systems on %s" % pair.primary)
    if selected_list is None:
        log_error("Selecting a file system(s) did not succeed.")
        print "Process canceled."
        raise Exceptions.NoExecution()

    selected_fs_list = [fs_list[idx - 1] for idx in selected_list]

    # Select a policy for the file system(s)
    policy_id = select_policy(model, is_all=pair.is_all)

    # Build StoragePolicy.
    # Some additional information is required for some policy types.
    if policy_id == FailoverConfig.StoragePolicy.AUTOUSAGE:
        # Ask 'margin' capacity
        margin = input_margin_capacity()

        # Set the policy for the specified file systems
        for fs in selected_fs_list:
            policy = FailoverConfig.StoragePolicy(fs.fileSystemName, policy=policy_id, margin=margin)
            # Delete and add the policy from/to storage configuration
            stconf.delete(pid, target_fs=fs.fileSystemName)
            stconf.add(pid, policy=policy)

    elif policy_id == FailoverConfig.StoragePolicy.AUTOSPEC:
        # Ask a fixed capacity
        capacity = input_specified_capacity()

        # Set the policy for the specified file systems
        for fs in selected_fs_list:
            policy = FailoverConfig.StoragePolicy(fs.fileSystemName, policy=policy_id, spec=capacity)
            # Delete and add the policy from/to storage configuration
            stconf.delete(pid, target_fs=fs.fileSystemName)
            stconf.add(pid, policy=policy)
    elif policy_id == FailoverConfig.StoragePolicy.MANUAL:
        # Ask manual configuration for each file system
        for fs in selected_fs_list:
            # Check if Active Migration is used on the file system
            is_actmig = 'ActMig' in fs.usage

            # Configure a manual policy
            policy = input_manual_storage_configuration(fs.fileSystemName, model, is_actmig, pair, config, ssh_config)

            # Delete and add the policy from/to storage configuration
            stconf.delete(pid, target_fs=fs.fileSystemName)
            stconf.add(pid, policy=policy)
    else:
        # Other than above. No additional information is required.
        policy = FailoverConfig.StoragePolicy(fs.fileSystemName, policy=policy_id)

        stconf.delete(pid, target_fs=fs.fileSystemName)
        stconf.add(pid, policy=policy)

def delete_policy(pid, pair, stconf):
    """ Delete a policy.

    Displays the current defined policies for the specified pair and let
    the admin select one of them to be deleted.

    Args:
        pid (int): Pair ID of the pair.
        pair (FailoverConfig.Pair): Pair instance of the pair.
        stconf (FailoverConfig.StoragePolicyConfig): Storage policy configuration.

    Returns:
        None: Nothing is returned. Just updates stconf.
    """

    print
    print "----- Delete a policy -----"

    # Get the current defined policies for the pair for selection.
    policy_list = stconf.get_list(pair_id=pid)
    if len(policy_list) == 0:
        log_error("No policies are defined for a pair. pid=%s, pair: %s" % (pid, pair))
        print
        print "Wrong state - no policies are defined for the pair. Please delete the pair, and add again to fix this state."
        raise Exceptions.NoExecution()

    # Remove default policy which cannot be deleted.
    policy_list_for_menu = filter(lambda e: e[1] != FailoverConfig.StoragePolicy.FSDEF, policy_list)

    # Build a menu to select a policy to be deleted.
    policy_menu = []
    for i, (notused, fs, policy) in enumerate(policy_list_for_menu):
        idx = i + 1
        policy_menu.append((idx, "%s - %s" % (fs, policy.to_display_string())))

    if len(policy_menu) == 0:
        print
        print "*** No policies are defined ***"
        raise Exceptions.NoExecution()

    # Select a policy to delete
    print
    print ">>> Specify a policy to delete."
    selected_list = Utility.select_items("Select a policy to delete", policy_menu, title="Defined policies", onlyone=True)
    if selected_list is None:
        log_warning("Selecting a policy did not succeed.")
        print "Process canceled."
        raise Exceptions.NoExecution()
    selected = selected_list[0]
    selected_fs = policy_list_for_menu[selected - 1][1]
    log_info("%s is selected to be deleted." % selected_fs)

    # Delete the policy
    print
    print "Deleting the policy ..."
    stconf.delete(pid, target_fs=selected_fs)

def display_policies(pid, pair, stconf):
    # Same as display_pairs(), but just displays only this pair.
    print
    print "----- Display storage policies -----"

    # There is at least one policy (default policy)
    print
    setup_common.display_pair_line('#', 'Primary', 'Secondary', 'File system', 'Storage policy')
    print "------------------------------------------------------------------------"
    policy_list = stconf.get_list(pair_id=pid)
    setup_common.display_pair(1, pair, policy_list)
    print "------------------------------------------------------------------------"

def display_file_systems(pair, config, ssh_config):
    """ Display file systems on the primary node of the pair.

    Display a list of file systems on the primary node. The list contains
    file system name and capacity for each file system.

    Args:
        pair (FailoverConfig.Pair): Pair instance of the pair.
        config (Config.Config): Configuration of the scripts.
        ssh_config (Config.SSHConfig): SSH configuration.

    Returns:
        None: Nothing is returned.
    """
    print
    print "----- Display file systems on the primary node -----"
    # Get file systems on the primary node.
    fs_list = getinfo_common.get_remote_file_systems(pair.primary, config, ssh_config)
    if len(fs_list) == 0:
        print
        print "*** No file systems found on the primary node ***"
        log_info("No file systems found on %s" % pair.primary)
        raise Exceptions.NoExecution()

    fs_menu = []
    for i, fs in enumerate(fs_list):
        idx = i + 1
        fs_menu.append((idx, "%s - %s GB" % (fs.fileSystemName, fs.fileSystemSizeInGigaByte)))

    print
    Utility.display_menu(fs_menu, "File systems and capacity on the primary node")

def configure_storage_policy(pairconf, stconf, config, ssh_config):
    """ Configure a storage policy.

    User selects either of:
    1. Change default policy
    2. Define a policy
    3. Delete a policy
    4. Display defined policies
    5. Display file systems on the primary node

    And updates storage configuration accordingly.

    Args:
        pairconf (FailoverConfig.PairConfig): Pair configuration.
        stconf (FailoverConfig.StoragePolicyConfig): Storage policy configuration.
        config (Config.Config): The script configuration.
        ssh_config (Config.SSHConfig): SSH configuration.

    Returns:
        None: Returns nothing. Update stconf.
    """

    # Select a pair to configure storage policy
    pid = setup_common.select_pair(pairconf)
    if pid is None:
        log_warning("No pairs are selected.")
        return

    pair_list = pairconf.get_list(pid)
    pair = pair_list[0][1]
    log_info("Selected pair: %s" % pair)

    # Configuration menu
    OP_CHANGE_DEFAULT = 1
    OP_DEFINE_POLICY = 2
    OP_DELETE_POLICY = 3
    OP_DISPLAY_POLICY = 4
    OP_DISPLAY_FILESYSTEM = 5
    OP_EXIT = 0
    menu = [
        (OP_CHANGE_DEFAULT, "Change default policy"),
        (OP_DEFINE_POLICY, "Define a policy for a file system"),
        (OP_DELETE_POLICY, "Delete a policy"),
        (OP_DISPLAY_POLICY, "Display defined policies"),
        (OP_DISPLAY_FILESYSTEM, "Display file systems on the primary node"),
        (OP_EXIT, "Finish")
    ]
    while True:
        selected = setup_common.select_operation(menu)
        try:
            if selected == OP_CHANGE_DEFAULT:
                change_default_policy(pid, pair, stconf, config, ssh_config)
                save_configuration(stconf)
            elif selected == OP_DEFINE_POLICY:
                define_policy(pid, pair, stconf, config, ssh_config)
                save_configuration(stconf)
            elif selected == OP_DELETE_POLICY:
                delete_policy(pid, pair, stconf)
                save_configuration(stconf)
            elif selected == OP_DISPLAY_POLICY:
                display_policies(pid, pair, stconf)
            elif selected == OP_DISPLAY_FILESYSTEM:
                display_file_systems(pair, config, ssh_config)
            elif selected == OP_EXIT:
                log_info("Exit from configuration menu.")
                break
        except Exceptions.NoExecution:
            print "Going back to configuration menu ..."

        print
        notused = raw_input(">>> Press Enter to go back to configuration menu.")

def save_configuration(stconf):
    # Save the configuration
    print
    print "Saving the configuration ..."
    stconf.save()

def main():
    # Enable Ctrl-C interruption
    Utility.enable_interrupt()

    # Load configs
    config = Config.Config()

    # Setup logging
    log_file_path = os.path.join(config.log_dir, config.common_log_file)
    setup_rotation_logger(__name__, config.log_level, log_file_path, config.log_size, config.log_gen)

    log_info("----------------------- %s --------------------------" % PROGRAM)

    # Load the existing pair configuration
    log_info("Getting list of pair and storage configuration.")
    try:
        confgroup_list = FailoverConfig.list_configuration_group(config.fo_conf_dir)
    except FailoverConfig.ConfigurationError, e:
        log_error("Failed to count number of configurations due to an configuration error. error=%s" % str(e))
        raise Exceptions.ConfigurationError(str(e))

    # Select a configuration group
    confgroup = setup_common.select_configuration_group(confgroup_list)

    # Load pair configuration and storage policy configuration
    pairconf = FailoverConfig.PairConfig(config.fo_conf_dir, confgroup)
    stconf = FailoverConfig.StoragePolicyConfig(config.fo_conf_dir, pairconf.confgroup)

    # Setup SSHConfig
    ssh_config = Config.SSHConfig(config.sys_conf_dir, pairconf, confgroup)

    # Main menu
    OP_DISPLAY = 1
    OP_CONFIG = 2
    OP_EXIT = 0
    main_menu = [
        (OP_DISPLAY, "Display pairs and storage policies"),
        (OP_CONFIG, "Configure storage policy"),
        (OP_EXIT, "Exit")
    ]

    while True:
        # Select an operation
        selected = setup_common.select_operation(main_menu)

        try:
            # Execute the selected operation
            if selected == OP_DISPLAY:
                print
                print "----- Display pairs -----"
                setup_common.display_pairs(pairconf, stconf)
            elif selected == OP_CONFIG:
                print
                print "----- Configure storage policy -----"
                configure_storage_policy(pairconf, stconf, config, ssh_config)
            elif selected == OP_EXIT:
                log_info("Exit selected.")
                break
        except Exceptions.NoExecution:
            print "Going back to the main menu ..."

        print
        notused = raw_input(">>> Press Enter to go back to the main menu.")

if __name__ == '__main__':
    try:
        main()
    except Exceptions.NoExecution:
        sys.exit(0)
    except KeyboardInterrupt:
        print "Interrupted."
        log_warning("Interrupted.")
        sys.exit(2)
    except Config.BaseError, e:
        print "Unable to load script configuration: %s" % str(e)
        sys.exit(10)
    except Exceptions.ConfigurationError, e:
        log_error("Unexpected configuration found: %s" % str(e))
        sys.exit(20)
    except Exceptions.ConnectionError, e:
        log_error("A connection error occurred: %s" % str(e))
        sys.exit(70)
    except Exceptions.OperationError, e:
        log_error("An operation error occurred: %s" % str(e))
        sys.exit(90)
    except EnvironmentError, e:
        log_exception("A system error occurred: %s", e)
        sys.exit(99)
    except Exception, e:
        log_exception("An internal error occurred: %s", e)
        sys.exit(99)

    sys.exit(0)
